import numpy as np
from skimage import exposure


def to_uint8(im, is_linear=True):
    """将栅格转为UINT8

    Args:
        im (np.ndarray): 栅格图像
        is_linear (bool, optional): 是否使用2%线性拉伸，默认为True

    Returns:
        np.ndarray: UINT8的栅格图像
    """

    # 2% linear stretch
    def _two_percent_linear(image, max_out=255, min_out=0):
        def _gray_process(gray, maxout=max_out, minout=min_out):
            # Get the corresponding gray level at 98% in the histogram.
            high_value = np.percentile(gray, 98)
            low_value = np.percentile(gray, 2)
            truncated_gray = np.clip(gray, a_min=low_value, a_max=high_value)
            processed_gray = (
                (truncated_gray - low_value) / (high_value - low_value)
            ) * (maxout - minout)
            return np.uint8(processed_gray)

        if len(image.shape) == 3:
            processes = []
            for b in range(image.shape[-1]):
                processes.append(_gray_process(image[:, :, b]))
            result = np.stack(processes, axis=2)
        else:  # if len(image.shape) == 2
            result = _gray_process(image)
        return np.uint8(result)

    # Simple image standardization
    def _sample_norm(image):
        stretches = []
        if len(image.shape) == 3:
            for b in range(image.shape[-1]):
                stretched = exposure.equalize_hist(image[:, :, b])
                stretched /= float(np.max(stretched))
                stretches.append(stretched)
            stretched_img = np.stack(stretches, axis=2)
        else:  # if len(image.shape) == 2
            stretched_img = exposure.equalize_hist(image)
        return np.uint8(stretched_img * 255)

    dtype = im.dtype.name
    if dtype != "uint8":
        im = _sample_norm(im)
    if is_linear:
        im = _two_percent_linear(im)
    return im



import os.path as osp
from glob import glob


PATH = osp.dirname(osp.abspath(__file__))


def get_file_name(filr_path: str) -> str:
    return filr_path.replace("\\", "/").split("/")[-1].split(".")[0]


SATELLITES = {}
satellite_dir = osp.join(PATH, "satellite")
yamls = glob(osp.join(satellite_dir, "*/*.yaml"))
for yaml_file in yamls:
    name = get_file_name(yaml_file)
    SATELLITES[name] = yaml_file



"""公式中的索引来自波段的description，目前仅支持四则运算"""

RSINDEXS = {
    "NDVI": "(NIR - Red) / (NIR + Red)",  # 归一化植被指数
    "EVI": "2.5 * (NIR - Red) / (NIR + 6 * Red - 7.5 * Blue + 1)",  # 增强型植被指数
    "RVI": "NIR / Red",  # 比值植被指数
    "GNDVI": "(NIR - Green) / (NIR + Green)",  # 绿色归一化差值植被指数
    "DVI": "NIR - Red",  # 差值植被指数
    "NDWI": "(Green - NIR) / (Green + NIR)",  # 归一化水体指数
    "GCVI": "(NIR / Green) - 1",  # 叶绿素指数
}



from typing import Union, Optional, Iterable, List, Any
import yaml
import pandas as pd
from .error import AttrNotFindError


class Band:
    __slots__ = "_attrs"

    def __init__(
        self,
        tag: str,
        order: int,
        resolution: Union[int, float],
        wavelength: Optional[Union[float, Iterable[float]]] = None,
        description: Optional[str] = None,
    ) -> None:
        """波段类

        Args:
            tag (str): 该波段的标签
            order (int): 该波段的索引偏移
            resolution (Union[int, float]): 该波段图像的分辨率
            wavelength (Optional[Union[float, Iterable[float]]], optional): 该波段的波长，默认为None
            description (Optional[str], optional): 该波段的相关描述，默认为None
        """
        self._attrs = {
            "tag": tag,
            "order": order,
            "resolution": resolution,
            "wavelength": wavelength,
            "description": description,
        }

    def get(self, type: str = "description") -> Any:
        """获取波段的某个属性

        Args:
            type (str, optional): 需要获取的属性，默认为"description"

        Raises:
            AttrNotFindError: 该属性不存在

        Returns:
            Any: 属性值
        """
        if type not in self._attrs.keys():
            raise AttrNotFindError("Cannt find {} in band attrs.".format(type))
        return self._attrs[type]

    def set(self, value: Any, type: str = "description") -> None:
        """设置波段的某个属性

        Args:
            value (Any): 属性值
            type (str, optional): 需要设置的属性，默认为"description"

        Raises:
            AttrNotFindError: 该属性不存在
        """
        if type not in self._attrs.keys():
            raise AttrNotFindError("Cannt find {} in band attrs.".format(type))
        self._attrs[type] = value

    def summay(self) -> None:
        """展示波段的属性"""
        print(pd.DataFrame(data=[self._attrs]))


class BandList:
    def __init__(self, bands: Iterable[Band]) -> None:
        """波段列表类

        Args:
            bands (Iterable[Band]): 一个包含多个波段的序列
        """
        self.band_list = tuple(bands)

    def __len__(self) -> int:
        return len(self.band_list)

    def find(self, value: Any, type: str = "description") -> List[Band]:
        """从某种属性中查找满足条件的波段

        Args:
            value (Any): 查找的值
            type (str, optional): 查找的属性，默认为"description"

        Raises:
            AttrNotFindError: 该属性不存在或波段列表中没有波段

        Returns:
            List[Band]: 满足条件的波段组成的列表
        """
        if self.__len__ == 0 or type not in self.band_list[0]._attrs:
            raise AttrNotFindError("Cannt find {} in band list.".format(type))
        result = []
        for band in self.band_list:
            if type != "wavelength":
                if value == band.get(type):
                    result.append(band)
            else:
                wl = band.get(type)
                if isinstance(wl, (int, float)):
                    if value == wl:
                        result.append(band)
                else:
                    if value >= wl[0] and value <= wl[1]:
                        result.append(band)
        return result

    def summay(self) -> None:
        """展示所有波段的属性总览"""
        dfs = []
        for band in self.band_list:
            dfs.append(band._attrs)
        print(pd.DataFrame(data=dfs))


def creat_band_list_from_config(yaml_path: str) -> BandList:
    """通过配置文件创建波段列表

    Args:
        yaml_path (str): 配置文件的路径

    Returns:
        BandList: 波段列表
    """
    band_list = []
    with open(yaml_path, "r", encoding="utf-8") as f:
        cfg = yaml.load(f.read(), Loader=yaml.Loader)
        for ib in cfg["bands"]:
            b = cfg["bands"][ib]
            band_list.append(
                Band(
                    b["tag"],
                    b["order"],
                    b["resolution"],
                    b["wavelength"],
                    b["description"],
                )
            )
    return BandList(band_list)



import os.path as osp
from math import ceil
from typing import Any, Iterator, Optional, Tuple
import re
import numpy as np
from osgeo import gdal
from .functions import RSINDEXS
from .utils import to_uint8
from .error import *
from ..config import SATELLITES
from ..band import creat_band_list_from_config


class SatelliteReader:
    def __init__(self, sensing: str = "ElectronicMap") -> None:
        """卫星图像加载器

        Args:
            sensing (str, optional): 传感器类型或或自定义配置，默认为"ElectronicMap"
        """
        if osp.exists(sensing) and sensing.split(".")[-1] in ("yaml", "yml"):
            config_file = sensing
        elif sensing in SATELLITES.keys():
            config_file = SATELLITES[sensing]
        else:
            SatelliteNotFindError("Cannt find {} in SATELLITES.".format(sensing))
        self._band_list = creat_band_list_from_config(config_file)
        self._src = None
        self._width = None
        self._height = None
        self._bands = None

    @classmethod
    def support_sensing(cls) -> Tuple[str, ...]:
        """支持的传感器类型

        Returns:
            Tuple[str, ...]: 支持的传感器类型
        """
        return tuple(SATELLITES.keys())

    @classmethod
    def support_index(cls) -> Tuple[str, ...]:
        """支持的波段计算公式

        Returns:
            Tuple[str, ...]: 支持的波段计算公式
        """
        return tuple(RSINDEXS.keys())

    def summay(self) -> None:
        """展示所有波段的属性总览"""
        self._band_list.summay()

    def open(self, img_path: str) -> None:
        """打开一个栅格图像

        Args:
            img_path (str): 图像路径

        Raises:
            OpenFileError: 图像路径不存在
            MismatchError: 图像波段数与传感器波段数不一致
        """
        self._src = gdal.Open(img_path)
        if self._src is None:
            raise OpenFileError("Open file failed.")
        self._width = self._src.RasterXSize
        self._height = self._src.RasterYSize
        self._bands = self._src.RasterCount
        if self._bands != len(self._band_list):
            raise MismatchError(
                "The lenght of data and source bands should be equal, but {} is not equal to {}.".format(
                    self._bands, len(self._band_list)
                )
            )

    def get_band(
        self,
        value: Any,
        type: str = "description",
        window: Optional[Tuple[int, int, int, int]] = None,
    ) -> np.ndarray:
        """获取一个波段的值，若索引出的波段大于两个，也只将返回第一个波段的值

        Args:
            value (Any): 查找的值
            type (str, optional): 查找的属性，默认为"description"
            window (Optional[Tuple[int, int, int, int]], optional): 窗口位置，默认为None

        Raises:
            OpenFileError: 图像未打开

        Returns:
            np.ndarray: 该波段的值
        """
        if self._src is None:
            raise OpenFileError("Please read image file first!")
        order = self._band_list.find(value, type)[0].get("order")
        if window is None:
            return self._src.GetRasterBand(order).ReadAsArray()
        else:
            return self._src.GetRasterBand(order).ReadAsArray(
                xoff=window[0], yoff=window[1], win_xsize=window[2], win_ysize=window[3]
            )

    def get_RGB(
        self, to_u8: bool = True, window: Optional[Tuple[int, int, int, int]] = None
    ) -> np.ndarray:
        """获取栅格的真彩色图像

        Args:
            to_u8 (bool, optional): 是否转为UINT8，默认为True.
            window (Optional[Tuple[int, int, int, int]], optional): 窗口位置，默认为None

        Returns:
            np.ndarray: 该栅格的真彩色图像
        """
        ima = np.stack(
            [
                self.get_band("Red", window=window),
                self.get_band("Green", window=window),
                self.get_band("Blue", window=window),
            ],
            axis=-1,
        )
        if to_u8:
            ima = to_uint8(ima)
        return ima

    def sample_band_compute(
        self,
        mode: str = "NDVI",
        eps: float = 2.2204e-16,
        window: Optional[Tuple[int, int, int, int]] = None,
    ) -> np.ndarray:
        """简单的波段计算

        Args:
            mode (str, optional): 需要计算的指数，默认为"NDVI"
            eps (float, optional): 避免分母为零，默认为2.2204e-16.
            window (Optional[Tuple[int, int, int, int]], optional): 窗口位置，默认为None

        Raises:
            FunctionNotFindError: 图像未打开

        Returns:
            np.ndarray: 波段计算的结果
        """
        if mode not in RSINDEXS.keys():
            raise FunctionNotFindError("Cannt find {} in functions.".format(mode))
        func = RSINDEXS[mode]
        values = re.split("\+|-|\*|/", func)
        for value in list(set(values)):
            value = value.strip().replace("(", "").replace(")", "")
            try:
                _ = float(value)
            except:
                exec(value + " = self.get_band(value, window=window) + eps")
        return eval(func)

    def _get_window(
        self, start_loc: Tuple[int, int], block_size: Tuple[int, int]
    ) -> Tuple[int, int, int, int]:
        if len(start_loc) != 2 or len(block_size) != 2:
            raise ValueError("The length start_loc/block_size must be 2.")
        if self._width is None or self._height is None:
            raise OpenFileError("Please read image file first!")
        xoff, yoff = start_loc
        xsize, ysize = block_size
        if (xoff < 0 or xoff > self._width) or (yoff < 0 or yoff > self._height):
            raise ValueError(
                "start_loc must be within [0-{0}, 0-{1}].".format(
                    str(self._width), str(self._height)
                )
            )
        if xoff + xsize > self._width:
            xsize = self._width - xoff
        if yoff + ysize > self._height:
            ysize = self._height - yoff
        return (int(xoff), int(yoff), int(xsize), int(ysize))

    def get_iter(
        self, block_size: Tuple[int, int] = (512, 512), mode: str = "ALL"
    ) -> Iterator[np.ndarray]:
        """迭代返回固定大小的图像块

        Args:
            block_size (Tuple[int, int], optional): 图像块大小，默认为(512, 512)
            mode (str, optional): 图像模式，默认为"ALL"

        Raises:
            OpenFileError: 图像未打开

        Yields:
            Iterator[np.ndarray]: 图像块的迭代器
        """
        if self._width is None or self._height is None:
            raise OpenFileError("Please read image file first!")
        rows = ceil(self._height / block_size[0])
        cols = ceil(self._width / block_size[1])
        for r in range(rows):
            for c in range(cols):
                loc_start = (int(c * block_size[1]), int(r * block_size[0]))
                xoff, yoff, xsize, ysize = self._get_window(loc_start, block_size)
                if mode == "RGB":
                    ima = self.getRGB(window=(xoff, yoff, xsize, ysize))
                elif mode in RSINDEXS.keys():
                    ima = self.sample_band_compute(
                        mode, window=(xoff, yoff, xsize, ysize)
                    )
                else:
                    if self._src is None:
                        raise OpenFileError("Please read image file first!")
                    ima = self._src.ReadAsArray(
                        xoff=xoff, yoff=yoff, xsize=xsize, ysize=ysize
                    )
                h, w = ima.shape[:2] if len(ima.shape) == 3 else ima.shape
                c = ima.shape[-1] if len(ima.shape) == 3 else 1
                if c != 1:
                    tmp = np.zeros((block_size[0], block_size[1], c), dtype=ima.dtype)
                    tmp[:h, :w, :] = ima
                else:
                    tmp = np.zeros((block_size[0], block_size[1]), dtype=ima.dtype)
                    tmp[:h, :w] = ima
                yield tmp



import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    LONG_DESCRIPTION = fh.read()

with open("requirements.txt", "r") as fin:
    REQUIRED_PACKAGES = fin.read()

setuptools.setup(
    name="sdrl",
    version="0.0.6",
    author="geoyee",
    author_email="geoyee@yeah.net",
    description="Satellite Data Read Library",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/geoyee/SDRL",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=REQUIRED_PACKAGES,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)
